import re
import stat
import time
from mimetypes import types_map
import os
from . import filesys
from rs4 import producers
from rs4.termcolor import tc
from aquests.protocols.http import http_date
from aquests.protocols.http.http_util import *
from hashlib import md5
from ..corequest import tasks
from aquests.athreads import trigger

IF_MODIFIED_SINCE = re.compile (
	'([^;]+)((; length=([0-9]+)$)|$)',
	re.IGNORECASE
	)

IF_NONE_MATCH = re.compile (
	'"?(.+?)(?:$|")',
	re.IGNORECASE
	)

USER_AGENT = re.compile ('User-Agent: (.*)', re.IGNORECASE)

CONTENT_TYPE = re.compile (
	r"Content-Type: ([^;]+)((; boundary=([A-Za-z0-9'\(\)+_,./:=?-]+)$)|$)",
	re.IGNORECASE
	)

def get_re_match (head_reg, value):
	m = head_reg.match (value)
	if m and m.end() == len (value):
		return m
	return ''

class MemCache:
	def __init__ (self):
		self.d = {}

	def read (self, path, file, etag):
		cached = self.d.get (path)
		if cached is None or cached [0] != etag:
			with open (file, "rb") as f:
				self.d [path] = (etag, f.read ())
		return self.d [path][1]


class Handler:
	directory_defaults = [
		'index.htm',
		'index.html'
		]
	_in_thread = False

	def __init__ (self, wasc, pathmap = {}, max_ages = None, alt_handlers = []):
		self.wasc = wasc

		self.memcache = MemCache ()
		self.max_ages = []
		if max_ages:
			self.max_ages = [(k, v) for k, v in max_ages.items ()]
			self.max_ages.sort (key = lambda x: len (x [0]), reverse = True)

		self.alt_handlers = alt_handlers
		self.permit_cache = {}
		self.filesystem = filesys.mapped_filesystem ()
		for k, v in list(pathmap.items ()):
			self.add_route (k, v)

	def close (self):
		for h in self.alt_handlers:
			try: h.close ()
			except AttributeError: pass

	def match (self, request):
		return 1

	def add_route (self, alias, option):
		self.wasc.logger ('server', 'directory {} mounted to {}'.format (option, tc.white (alias or '/')), 'info')
		self.filesystem.add_map (alias, option)

	def handle_alternative (self, request):
		for h in self.alt_handlers:
			if h.match (request):
				try:
					h.handle_request (request)
				except:
					self.wasc.logger.trace("server")
					try: request.response.error (500)
					except: pass
				return
		request.response.error (404)

	def isprohibited (self, request, path):
		dirname = os.path.split(path) [0]
		permission = self.filesystem.get_permission (dirname)
		if not permission:
			return False
		if not self.wasc.authorizer.has_permission (request, permission):
			return True
		return False

	def is_etag_matched (self, request, header_name, etag):
		hval = request.get_header (header_name)
		if not hval:
			return
		match = get_re_match (IF_NONE_MATCH, hval)
		if not match:
			return
		return etag == match.group (1) and 'matched' or 'unmatched'

	def is_modified (self, request, header_name, mtime, file_length):
		ims_h = request.get_header (header_name)
		if not ims_h:
			return
		match = get_re_match (IF_MODIFIED_SINCE, ims_h)
		if not match:
			return

		length = match.group (4)
		if length:
			try:
				length = int(length)
				if length != file_length:
					return 'modified'
			except:
				pass

		try:
			mtime2 = http_date.parse_http_date (match.group (1))
		except:
			return

		if mtime > mtime2:
			return 'modified'
		else:
			return 'unmodified'

	def handle_request (self, request):
		if request.command not in ('get', 'head'):
			self.handle_alternative (request)
			return

		path, params, query, fragment = request.split_uri()
		if '%' in path:
			path = unquote (path)

		# return redirect
		if path == "":
			request.response['Location'] = '/'
			request.response.error (301)
			return

		if path.find ("./") != -1 or path.find ("//") != -1:
			request.response.error (404)
			return

		while path and path [0] == '/':
			path = path[1:]

		if self.filesystem.isdir (path):
			if path and path[-1] != '/':
				request.response['Location'] = '/%s/' % path
				request.response.error (301)
				return

			found = False
			for default in self.directory_defaults:
				p = path + default
				if self.filesystem.isfile (p):
					path = p
					found = True
					break

			if not found:
				self.handle_alternative (request)
				return

		elif not self.filesystem.isfile (path):
			return self.handle_alternative (request)

		if path and path [-1] == "/":
			return request.response.error (404)

		if self.isprohibited (request, path):
			return

		try:
			current_stat = self.filesystem.stat (path)
			mtime = current_stat [stat.ST_MTIME]
			file_length = current_stat [stat.ST_SIZE]
		except:
			self.handle_alternative (request)
			return

		etag = self.make_etag (file_length, mtime)
		if self.is_etag_matched (request, 'if-none-match', etag) == 'matched' or self.is_modified (request, "if-modified-since", mtime, file_length) == 'unmodified':
			self.set_cache_control (request, path, mtime, etag)
			request.response.start (304)
			request.response.done()
			return

		if self.is_etag_matched (request, 'if-match', etag) == 'unmatched' or self.is_modified (request, "if-unmodified-since", mtime, file_length) == 'modified':
			request.response.start (412)
			request.response.done()
			return

		range_ = request.get_header ('range')
		if range_:
			if self.is_etag_matched (request, 'if-range', etag) == 'unmatched':
				range_ = None # ignore range
			else:
				try:
					rg_start, rg_end = parse_range (range_, file_length)
				except:
					request.response.start (416)
					request.response.done()
					return

		self.set_cache_control (request, path, mtime, etag)
		self.set_content_type (path, request)

		if range_:
			request.response ['Content-Range'] = 'bytes {}-{}/{}'.format (rg_start, rg_end, file_length)
			request.response ['Content-Length'] = (rg_end - rg_start) + 1
			offset, limit = rg_start, (rg_end - rg_start) + 1
			request.response.start (206)

		else:
			request.response ['Content-Length'] = file_length
			offset, limit = 0, file_length

		if request.command == 'get' and limit:
			# if head, don't send contents
			if file_length < 4096:
				request.response.push (
					self.memcache.read (path, self.filesystem.translate (path), etag) [offset:limit]
				)
			else:
				request.response.push (producers.file_producer (
					self.filesystem.open (path, 'rb'), proxsize = file_length, offset = offset, limit = limit)
				)

		if self._in_thread:
			trigger.wakeup (lambda p = request.response: (p.done (),))
		else:
			request.response.done()

	def set_cache_control (self, request, path, mtime, etag):
		max_age = self.max_ages and self.get_max_age (path) or 0
		if request.version == "1.0":
			request.response ['Last-Modified'] = http_date.build_http_date (mtime)
			if max_age:
				request.response ['Expires'] = http_date.build_http_date (mtime + max_age)
		else:
			request.response ['Etag'] = '"' + etag + '"'
			if max_age:
				request.response ['Cache-Control'] = "max-age=%d" % max_age

	def get_max_age (self, path):
		path = not path and "/" or "/" + path
		for prefix, value in self.max_ages:
			if path.startswith (prefix):
				return value

	def make_etag (self, file_length, mtime):
		return md5 (("%d:%d" % (file_length, int (mtime))).encode ("utf8")).hexdigest()

	def set_content_type (self, path, request):
		ext = get_extension (path).lower()
		request.response['Content-Type'] = types_map.get ("." + ext, 'application/octet-stream')

	def get_static_files (self):
		return StaticFiles (self.wasc, self.filesystem, self.max_ages, self.memcache)


class StaticFiles (Handler):
	_in_thread = True
	__name__ = 'StaticFiles' # django make error, why?
	def __init__ (self, wasc, filesystem, max_ages, memcache):
		self.wasc = wasc
		self.filesystem = filesystem
		self.max_ages = max_ages
		self.memcache = memcache

	def handle_alternative (self, request):
		request.response.error (404)

	def __call__ (self, request, uri):
		request._split_uri = (uri,) + request._split_uri [1:]
		self.handle_request (request)
		return tasks.Revoke ()




#-------------------------------------------------------
# Basic Blade Server Architecture R2
# Hans Roh (hansroh@gmail.com)
# 2014.4.24 Python 2.7 port (from 2.4)
#-------------------------------------------------------

HTTPS = True
import sys, time, os, threading
from .backbone import http_server
from skitai import lifetime
from warnings import warn
from .backbone import https_server
from skitai import start_was
from collections import deque
from aquests.athreads.fifo import await_fifo
from aquests.client import asynconnect
from aquests.client import socketpool
from aquests.athreads import threadlib, trigger
from rs4 import logger, confparse, pathtool
from aquests.dbapi import dbpool
from aquests.protocols.http import request_handler
from aquests.protocols import http2
from aquests.client import adns
from aquests.protocols import dns
if os.name == "nt":
	from rs4.psutil import schedule # cron like scheduler
from .handlers import proxy_handler, ipbl_handler, vhost_handler, forward_handler
import socket
import signal
import multiprocessing
from . import wsgiappservice
from .backbone import http_response
from .corequest import cachefs
from .corequest.httpbase import task, rcache
from .corequest.httpbase import cluster_manager as rcluster_manager
from .corequest.dbi import cluster_manager as dcluster_manager
from .corequest.dbi import task as dtask
import types
from .handlers.websocket import servers as websocekts
from .wastuff import selective_logger, triple_logger
from .corequest.pth import executors

class Loader:
	def __init__ (self, config = None, logpath = None, varpath = None, wasc = None, debug = 0):
		self.config = config
		self.instance_name = os.path.split (config)[-1][:-5]
		self.logpath = logpath
		self.varpath = varpath
		self.debug = debug
		self.num_worker = 1
		self.wasc = wasc or wsgiappservice.WAS
		self.ssl = False
		self.ctx = None
		self._exit_code = None
		self.wasc_kargs = {}
		self._fifo_switched = False
		self.config_logger (self.logpath)
		self.WAS_initialize ()
		self.configure ()
		self.WAS_finalize ()

	def configure (self):
		raise SystemExit("configure must be overided")

	def set_num_worker (self, num):
		if os.name == "nt":
			num = 1
		self.num_worker = num
		self.wasc.workers = num

	def WAS_initialize (self):
		self.wasc.log_base_path = self.logpath
		self.wasc.var_base_path = self.varpath
		self.wasc.register ("debug", self.debug)
		self.wasc.register ("clusters",  {})
		self.wasc.register ("clusters_for_distcall",  {})
		self.wasc.register ("workers", 1)
		self.wasc.register ("cachefs", None)
		websocekts.start_websocket (self.wasc)
		self.wasc.register ("websockets", websocekts.websocket_servers)
		self.switch_to_await_fifo ()

	def set_model_keys (self, keys):
		self.wasc._luwatcher.add (keys)

	def app_cycle (self, func):
		for h in self.wasc.httpserver.handlers:
			if isinstance (h, vhost_handler.Handler):
				for vhost in h.sites.values ():
					for apph in vhost.apps.modules.values ():
						getattr (apph, func) ()

	def WAS_finalize (self):
		global the_was

		self.wasc.register ("lifetime", lifetime)
		# internal connection should be http 1.1
		# because http2 single connection feature is useless on accessing internal resources
		# BUT we will use http2 when gRPC call, with just 1 stream per connection for speeding
		http2.MAX_HTTP2_CONCURRENT_STREAMS = 1
		request_handler.RequestHandler.FORCE_HTTP_11 = True
		self.app_cycle ('mounted')
		start_was (self.wasc, **self.wasc_kargs)

	def config_wasc (self, **kargs):
		self.wasc_kargs = kargs

	def config_dns (self, prefer_protocol = "tcp"):
		adns.init (self.wasc.logger.get ("server"), prefer_protocol = prefer_protocol)
		lifetime.maintern.sched (4.1, dns.pool.maintern)

	def config_executors (self, workers, zombie_timeout):
		self.wasc.register ("executors", executors.Executors (workers, zombie_timeout, self.wasc.logger.get ("server")))

	def config_cachefs (self, cache_dir = None, memmax = 0, diskmax = 0):
		self.wasc.cachefs = cachefs.CacheFileSystem (cache_dir, memmax, diskmax)

		socketfarm = socketpool.SocketPool (self.wasc.logger.get ("server"))
		self.wasc.clusters ["__socketpool__"] = socketfarm
		self.wasc.clusters_for_distcall ["__socketpool__"] = task.TaskCreator (socketfarm, self.wasc.logger.get ("server"), self.wasc.cachefs)

		dp = dbpool.DBPool (self.wasc.logger.get ("server"))
		self.wasc.clusters ["__dbpool__"] = dp
		self.wasc.clusters_for_distcall ["__dbpool__"] = dtask.TaskCreator (dp, self.wasc.logger.get ("server"))

	def switch_to_await_fifo (self):
		if self._fifo_switched: return
		asynconnect.AsynConnect.fifo_class = await_fifo
		asynconnect.AsynSSLConnect.fifo_class = await_fifo
		http_server.http_channel.fifo_class = await_fifo
		https_server.https_channel.fifo_class = await_fifo
		self._fifo_switched = True

	def config_rcache (self, maxobj = 1000):
		rcache.start_rcache (maxobj)
		self.wasc.register ("rcache", rcache.the_rcache)

	def config_certification (self, certfile, keyfile = None, pass_phrase = None):
		if not HTTPS:
			return
		self.ctx = https_server.init_context (certfile, keyfile, pass_phrase)
		self.ssl = True

	def config_forward_server (self, ip = "", port = 80, forward_to = 443):
		forward_server = http_server.http_server (ip or "", port, self.wasc.logger.get ("server"), self.wasc.logger.get ("request"))
		forward_server.zombie_timeout = 2
		forward_server.install_handler (forward_handler.Handler (self.wasc, forward_to))
		self.wasc.register ("forwardserver", forward_server)

	def config_webserver (self, port, ip = "", name = "", ssl = False, keep_alive = 10, network_timeout = 10, single_domain = None, thunks = [], quic = None, backlog = 100, multi_threaded = False, max_upload_size = 256000000):
		# maybe be configured	at first.
		if ssl and not HTTPS:
			raise SystemError("Can't start SSL Web Server")

		if not name:
			name = self.instance_name
		http_server.configure (name, network_timeout, keep_alive, multi_threaded, max_upload_size)

		if ssl and self.ctx is None:
			raise ValueError("SSL ctx not setup")
		if ssl:
			server_class = https_server.https_server
		else:
			server_class = http_server.http_server

		if self.ssl:
			httpserver = server_class (ip or "", port, self.ctx, quic, self.wasc.logger.get ("server"), self.wasc.logger.get ("request"))
		else:
			httpserver = server_class (ip or "", port, self.wasc.logger.get ("server"), self.wasc.logger.get ("request"))
		single_domain and httpserver.set_single_domain (single_domain)
		self.wasc.register ("httpserver", httpserver)
		# starting jobs before forking
		for thunk in thunks:
			thunk ()

		self.wasc.httpserver.serve (hasattr (self.wasc, "forwardserver") and self.wasc.forwardserver or None, backlog)
		self.fork ()

	def fork (self):
		#fork here
		_exit_code = self.wasc.httpserver.fork (self.num_worker)
		if _exit_code is not None:
			self.handle_exit_code (_exit_code)

	def handle_exit_code (self, _exit_code):
		self._exit_code = _exit_code
		try:
			os.wait ()
		except OSError:
			pass

	def get_exit_code (self):
		return self._exit_code

	def config_scheduler (self, conffile):
		if os.name == "nt":
			scheduler = schedule.Scheduler (self.wasc, conffile, self.wasc.logger.get ("server"))
			self.wasc.register ("scheduler", scheduler)

	def config_logger (self, path, media = None, log_off = [], file_loggings = None):
		if not media:
			if path is not None:
				media = ["file"]
			else:
				media = ["screen"]

		http_response.http_response.log_or_not = selective_logger.SelectiveLogger (log_off)
		self.wasc.register ("logger", triple_logger.Logger (media, path, file_loggings))

		if os.name != "nt" and path:
			def hUSR1 (signum, frame):
				self.wasc.logger.rotate ()
			signal.signal(signal.SIGUSR1, hUSR1)

	def config_threads (self, numthreads = 0):
		if numthreads > 0:
			trigger.start_trigger (self.wasc.logger.get ("server"))
			queue = threadlib.request_queue2 ()
			tpool = threadlib.thread_pool (queue, numthreads, self.wasc.logger.get ("server"))
			self.wasc.register ("queue",  queue)
			self.wasc.register ("threads", tpool)
			self.wasc.numthreads = numthreads

	def config_backends (self, backend_keep_alive, object_timeout, maintern_interval):
		rcluster_manager.ClusterManager.backend_keep_alive = backend_keep_alive
		rcluster_manager.ClusterManager.object_timeout = object_timeout
		rcluster_manager.ClusterManager.maintern_interval = maintern_interval

		dcluster_manager.ClusterManager.backend_keep_alive = backend_keep_alive
		dcluster_manager.ClusterManager.object_timeout = object_timeout
		dcluster_manager.ClusterManager.maintern_interval = maintern_interval

	def add_cluster (self, clustertype, clustername, clusterlist, ssl = 0, access = None, max_conns = 100):
		try:
			self.wasc.add_cluster (clustertype, clustername, clusterlist, ssl = ssl, access = access, max_conns = max_conns)
		except:
			self.wasc.logger.trace ("server")

	def install_handler_with_tuple (self, routes):
		if type (routes) is list:
			routes = {'default': routes}
		sroutes = []
		for domain in sorted (routes.keys ()): # must sort for lueatcher reservation
			sroutes.append ("@%s" % domain)
			for route, entity, pref in routes [domain]:
				appname = None
				if type (entity) is tuple:
					entity, appname = entity
				if entity.endswith (".py") or entity.endswith (".pyc"):
					entity = os.path.join (os.getcwd (), entity) [:-3]
					if entity [-1] == ".":
						entity = entity [:-1]
				sroutes.append (("%s=%s%s" % (route, entity, appname and ":" + appname or ""), pref))
		return sroutes

	def install_handler (self,
			routes = [],
			proxy = False,
			static_max_ages = None,
			blacklist_dir = None,
			unsecure_https = False,
			enable_apigateway = False,
			apigateway_authenticate = False,
			apigateway_realm = "API Gateway",
			apigateway_secret_key = None
		):

		if routes:
			if type (routes) is dict:
				routes = self.install_handler_with_tuple (routes)
			else:
				if type (routes) is not list:
					routes = [routes]
				if type (routes [0]) is tuple:
					routes = self.install_handler_with_tuple (routes)

		if blacklist_dir:
			self.wasc.add_handler (0, ipbl_handler.Handler, blacklist_dir)
		if proxy:
			self.wasc.add_handler (1, proxy_handler.Handler, self.wasc.clusters, self.wasc.cachefs, unsecure_https)

		vh = self.wasc.add_handler (
			1, vhost_handler.Handler,
			self.wasc.clusters, self.wasc.cachefs,
			static_max_ages,
			enable_apigateway, apigateway_authenticate, apigateway_realm, apigateway_secret_key
		)

		current_rule = "default"
		for line in routes:
			config = None
			if type (line) is tuple:
				line, pref = line
			line = line.strip ()
			if line.startswith (";") or line.startswith ("#"):
				continue
			elif line.startswith ("/"):
				reverse_proxing = vh.add_route (current_rule, line, pref)
			elif line:
				if line [0] == "@":
					line = line [1:].strip ()
				current_rule = line

	def run (self, timeout = 30):
		if self._exit_code is not None:
			self.close ()
			return self._exit_code # master process

		try:
			try:
				if "---memtrack" in sys.argv:
					self.wasc.logger ("server", "memory tracking enabled", "debug")
					lifetime.enable_memory_track ()
				if "---profile" in sys.argv:
					import cProfile
					self.wasc.logger ("server", "profiling enabled", "debug")
					cProfile.runctx ("lifetime.loop (timeout)", globals (), locals (), "profile.out")
				else:
					lifetime.loop (timeout)
			except:
				self.wasc.logger.trace ("server")
		finally:
			self.close ()

		return None # worker process

	def close (self):
		for attr, obj in list(self.wasc.objects.items ()):
			if attr == "logger":
				continue

			if attr == "clusters":
				self.wasc.logger ("server", "[info] cleanup %s" % attr)
				for name, cluster in obj.items ():
					cluster.cleanup ()
				continue

			if hasattr (obj, "cleanup"):
				try:
					self.wasc.logger ("server", "[info] cleanup %s" % attr)
					obj.cleanup ()
					del obj
				except:
					self.wasc.logger.trace ("server")

		self.app_cycle ('umounted')

		if os.name == "nt" or self.wasc.httpserver.worker_ident == "master":
			self.wasc.logger ("server", "[info] cleanup done, closing logger... bye")
			try:
				self.wasc.logger.close ()
				del self.wasc.logger
			except:
				pass



import time
from aquests.athreads import socket_map
from aquests.athreads import trigger
import threading
from ..httpbase import task, rcache
from rs4.attrdict import AttrDict
from aquests.dbapi import request
from rs4 import asyncore
from skitai import DB_PGSQL, DB_SQLITE3, DB_REDIS, DB_MONGODB
from skitai import REQFAIL, UNSENT, TIMEOUT, NETERR, NORMAL
from rs4.cbutil import tuple_cb

class OperationTimeout (Exception):
    pass

class RequestFailed (Exception):
    pass

class FailedRequest:
    def __init__ (self, expt):
        self.description = None
        self.data = None
        self.expt = expt

        self.code, self.msg = 500, str (expt)
        self.status_code, self.reason = self.code, self.msg

    def raise_for_status (self):
        if self.expt:
            raise self.expt
    reraise = raise_for_status


class Dispatcher (task.Dispatcher):
    def __init__ (self, cv, id, ident = None, filterfunc = None, callback = None):
        self._cv = cv
        self.id = id
        self.ident = ident
        self.filterfunc = filterfunc
        self.callback = callback
        self.creation_time = time.time ()
        self.status = UNSENT
        self.result = None

    def get_result (self):
        if not self.result:
            status = self.get_status ()
            if status == REQFAIL:
                self.result = task.Result (self.id, REQFAIL, FailedRequest (RequestFailed ("Request Failed")), self.ident)
            else:
                self.result = task.Result (self.id, TIMEOUT, FailedRequest (OperationTimeout ("Operation Timeout")), self.ident)
        return self.result

    def handle_result (self, request):
        if self.get_status () == TIMEOUT:
            # timeout, ignore
            return
        status = NORMAL
        if request.expt:
            reason = str (request.expt)
            if reason == "Operation Timeout":
                status = TIMEOUT
            elif reason == "Socket Panic":
                status = NETERR
        result = task.Result (self.id, status, request, self.ident)
        self.set_status (status, result)
        tuple_cb (self, self.callback)


#-----------------------------------------------------------
# Cluster Base Call
#-----------------------------------------------------------

class Task (task.Task):
    def __init__ (
            self,
            cluster,
            server,
            dbname,
            auth,
            dbtype,
            meta = None,
            use_cache = False,
            mapreduce = True,
            filter = None,
            callback = None,
            cache = None,
            timeout = 10,
            origin = None,
            logger = None
        ):

        self._server = server
        self._dbname = dbname
        self._auth = auth
        self._dbtype = dbtype
        if self._dbtype == DB_PGSQL:
            if self._dbname in (DB_SQLITE3, DB_REDIS):
                self._dbtype = self._dbname
                self._dbname = ""
            elif self._auth in (DB_MONGODB,):
                self._dbtype = self._auth
                self._auth = None
        self.set_defaults (cluster, meta, use_cache, mapreduce, filter, callback, cache, timeout, origin, logger)

    def _get_ident (self):
        cluster_name = self._cluster.get_name ()
        if cluster_name == "__dbpool__":
            _id = "%s/%s/%s" % (self._server, self._dbname, self._auth)
        else:
            _id = cluster_name
        _id += "/%s%s" % (
            ",".join (map (lambda x: str (x), self._cached_request_args)),
            self._mapreduce and "/M" or ""
            )
        return _id

    def _get_connection (self, id = None):
        if id is None: id = self._nodes.pop ()
        else: self._nodes = []

        if self._server:
            asyncon = self._cluster.get (self._server, self._dbname, self._auth, self._dbtype)
        else:
            asyncon = self._cluster.get (id)
        self._setup (asyncon)
        return asyncon

    def _build_request (self, method, params):
        # For Django QuerySet and SQLGen
        if hasattr (params [0], "query"):
            params = (str (params [0].query),) + params [1:]

        self._cached_request_args = (method, params) # backup for retry
        if self._use_cache and rcache.the_rcache:
            self._cached_result = rcache.the_rcache.get (self._get_ident (), self._use_cache)
            if self._cached_result is not None:
                self._cached_result.meta = self._meta
                self._callback and tuple_cb (self._cached_result, self._callback)
                return

        while self._avails ():
            asyncon = self._get_connection (None)
            rs = Dispatcher (
                self._cv, asyncon.address,
                ident = not self._mapreduce and self._get_ident () or None,
                filterfunc = self._filter,
                callback = self._collect
            )
            self._requests [rs] = asyncon
            req = request.Request (
                self._dbtype,
                self._server,
                self._dbname,
                self._auth,
                method, params,
                rs.handle_result,
                self._meta
            )

            # IMP: must modify, dbconn.ConnectProxy
            try:
                asyncon.execute (req, rs, self.handle_request_failed)
            except:
                self.handle_request_failed (rs, asyncon)
                continue

        self._request = req # sample for unitest
        trigger.wakeup ()
        return self


class Proxy:
    def __init__ (self, __class, cluster, *args, **kargs):
        self.__class = __class
        self.__cluster = cluster
        self.__args = args
        self.__kargs = kargs

    @property
    def dbtype (self):
        try:
            return self.__cluster.dbtype
        except AttributError:
            return self.__args [3]

    def __enter__ (self):
        return self

    def __exit__ (self, type, value, tb):
        pass

    def __getattr__ (self, name):
        self._method = name
        return self.__proceed

    def __proceed (self, *params):
        cdc = self.__class (self.__cluster, *self.__args, **self.__kargs)
        cdc._build_request (self._method, params)
        return cdc


class TaskCreator:
    def __init__ (self, cluster, logger):
        self.cluster = cluster
        self.logger = logger

    def __getattr__ (self, name):
        return getattr (self.cluster, name)

    def Server (self, server = None, dbname = None, auth = None, dbtype = None, meta = None, use_cache = True, mapreduce = False, filter = None, callback = None, cache = None, timeout = 10, caller = None):
        return Proxy (Task, self.cluster, server, dbname, auth, dbtype, meta, use_cache, mapreduce, filter, callback, cache, timeout, caller, self.logger)
